from dubhe_sdk.param.train_param import TrainParam
from dubhe_sdk.param.predict_param import PredictParam
from dubhe_sdk.param.inference_param import InferenceParam
from dubhe_sdk.service.ADCkafka import *
from dubhe_sdk.config import *
import typing
import logging


class ContextKey:
    """
    Key to get algorithm id from context, Str.
    """
    KEY_ALGORITHM_ID = "algorithm_id"

    """
    Key to get data paths from context, List.
    If can not to start with this epoch ( for example, no such saved model file in cache folder ), raise an exception
    """
    KEY_DATA_PATHS = "data_paths"

    """
    Key to get train code list from context, List.
    """
    KEY_CODE_LIST = "code_list"
    KEY_VERIFY_WAY = "verify_way"
    KEY_VERIFY_PATHS = "verify_paths"
    KEY_SPLIT_RATE = "split_rate"
    KEY_PRETRAINED_MODEL_PATH = "pretrained_model_path"
    KEY_PRETRAINED_EPOCH = "pretrained_epoch"
    KEY_TMP_DIR = "tmp_dir"
    KEY_OUTPUT_DIR = "output_dir"
    KEY_IMAGE_BASE_DIR = "image_base_dir"
    KEY_IMAGE_NAMES = "image_names"
    KEY_RESOURCE_ALLOCATION = "resource_allocation"
    ENV_INSTANCE_ID = "INSTANCE_ID"
    ENV_PLATFORM_TYPE = "PLATFORM_TYPE"
    ENV_RELY_MODEL_DATA = "RELY_MODEL_DATA"
    ENV_MODEL_PATH = "model_path"
    ENV_SPECIAL_PARAMS = "special_params"

    class ContextKeyError(TypeError):
        pass

    def __setattr__(self, key, value):
        if key in self.__dict__:
            raise self.ContextKeyError(f'Can not set ContextKey {key}')
        self.__dict__[key] = value


class Context(object):
    def __init__(self, logger):
        self.logger = logger
        self.local_logger = logging.getLogger("dubhe")
        self.local_logger.setLevel(logging.INFO)

        self.key_map = {}
        self.ds_map = {}
        self.out_dataset = None

    def get(self, key: str, default: str = ""):
        if key in self.key_map:
            return self.key_map[key]
        else:
            return default

    def log(self, msg, *args, **kwargs):
        self.local_logger.info(msg, *args, **kwargs)

        if self.logger is not None:
            pass

    def logProgress(self, progress: float):

        logstr = "Progress:{:.2%}, ".format(progress)
        self.local_logger.info(logstr)
        if PLATFORM_TYPE == AI_PLATFORM:
            progress_json = predict_progress_data(progress)
            send_kafka(MODEL_PROGRESS, progress_json, TOPIC_MODEL_STATUS)

        if self.logger is not None:
            # todo, send log to cloud
            pass

    def logProgressByBatch(self, epoch: int, batch_idx, batchsize, progress: float, Metrics: typing.Dict[str, typing.Union[float, int]]):

        logstr = "Epoch:{:d} [{:d}|{:d}], Progress:{:.2%}, ".format(epoch, batch_idx, batchsize, progress)
        for m in Metrics.items():
            logstr += "%s: %4s, "%(m[0],m[1])
        self.local_logger.info(logstr)
        if PLATFORM_TYPE == AI_PLATFORM:
            progress_json = train_progress_data(progress, epoch, Metrics)
            send_kafka(MODEL_PROGRESS, progress_json, TOPIC_MODEL_STATUS)

        if self.logger is not None:
            # todo, send log to cloud
            pass

    def logProgressByEpoch(self, epoch: int, progress: float, Metrics: typing.Dict[str, typing.Union[float, int]]):
        logstr = "Epoch:{:d}, Progress:{:.2%}, ".format(epoch, progress)
        for m in Metrics.items():
            logstr += "%s: %4s, "%(m[0],m[1])
        self.local_logger.info(logstr)

        if PLATFORM_TYPE == AI_PLATFORM:
            progress_json = train_progress_data(progress, epoch, Metrics)
            send_kafka(MODEL_PROGRESS, progress_json, TOPIC_MODEL_STATUS)

    def logResult(self, predict_json: typing.Dict):
        logstr = "Result data:{}, ".format(predict_json)
        self.local_logger.info(logstr)

        if PLATFORM_TYPE == AI_PLATFORM and TASK_TYPE != TASK_INFERENCE_TYPE:
            send_kafka(MODEL_RESULT, predict_json, TOPIC_MODEL_RESULT)

    def logResultDetails(self, img_path, ground_true, predict_code, conf, shape_type=None, position=None, special=None):
        predict_json = predict_details(img_path, ground_true, predict_code, conf, shape_type, position, special)
        logstr = "Result data:{}, ".format(predict_json)
        self.local_logger.info(logstr)

        if PLATFORM_TYPE == AI_PLATFORM and TASK_TYPE != TASK_INFERENCE_TYPE:
            send_kafka(MODEL_RESULT, predict_json, TOPIC_MODEL_RESULT)
        return predict_json



class ContextBuilder(object):
    def __init__(self, logger):
        self.ctx = Context(logger)

    def setInputParam(self, dict_params: dict):
        # 模型输出目录
        OUTPUT_MODEL_DIR = 'output/models'
        if dict_params['req_type'] == TASK_TRAIN_TYPE:
            inputParams = TrainParam()
            inputParams.__dict__ = dict_params
            self.ctx.key_map[ContextKey.KEY_DATA_PATHS] = inputParams.data_paths
            self.ctx.key_map[ContextKey.KEY_CODE_LIST] = inputParams.code_list
            self.ctx.key_map[ContextKey.KEY_SPLIT_RATE] = inputParams.split_rate
            self.ctx.key_map[ContextKey.KEY_OUTPUT_DIR] = OUTPUT_MODEL_DIR
            self.ctx.key_map[ContextKey.KEY_PRETRAINED_MODEL_PATH] = inputParams.pretrained_model_path
            self.ctx.key_map[ContextKey.KEY_PRETRAINED_EPOCH] = inputParams.pretrained_epoch

        elif dict_params['req_type'] == TASK_PREDICT_TYPE:
            inputParams = PredictParam()
            inputParams.__dict__ = dict_params
            self.ctx.key_map[ContextKey.KEY_DATA_PATHS] = inputParams.data_paths

        elif dict_params['req_type'] == TASK_INFERENCE_TYPE:
            inputParams = InferenceParam()
            inputParams.__dict__ = dict_params
            self.ctx.key_map[ContextKey.KEY_IMAGE_BASE_DIR] = inputParams.image_base_dir
            self.ctx.key_map[ContextKey.KEY_IMAGE_NAMES] = inputParams.image_names
            self.ctx.key_map[ContextKey.KEY_RESOURCE_ALLOCATION] = inputParams.resource_allocation

        return self

    def add(self, key: str, value):
        self.ctx.key_map[key] = value
        return self

    def build(self):
        return self.ctx








def train_decorator():
    def do_train(f):
        setattr(f, "_is_train_func", True)
        return f

    return do_train


def predict_decorator():
    def do_predict(f):
        setattr(f, "_is_predict_func", True)
        return f

    return do_predict

def inference_decorator():
    def do_inference(f):
        setattr(f, "_is_inference_func", True)
        return f

    return do_inference

def pipeline_decorator():
    def do_pipeline(f):
        setattr(f, "_is_pipeline_func", True)

        return f

    return do_pipeline


import typing


class OutDataset(object):
    def Add(self, Id: str):
        pass

    def setOutFile(self, Id: str, func: typing.Callable[[typing.IO], None], key: str = "default") -> bool:
        """
        Set output to dst dataset file
        For example in a preprocess predict method, will use this method to crop input image and save them to
        another dataset.
        If no dst dataset is assigned, the func will not be called and all annotations will be saved into input
        dataset.
        Only available in predict method.

        :return: is_saved: true if func is called and new file is saved to dst dataset or false
        """
        pass

    def setClassification(self, Id: str, clazz: str, key: str = "default"):
        pass

    def setDetection(self, Id: str, detection: typing.List, key: str = "default"):
        pass

    def setMetadata(self, Id: str, value: str, key: str = "default"):
        pass

    def flush(self, force: bool = False) -> bool:
        """
        flush and upload all output to cloud, so that we do not need to cache too much data here
        if run in local, only save in local
        """
        pass



import os
import sys

from flask import Flask, jsonify, request as flask_request
from concurrent.futures import ThreadPoolExecutor
from dubhe_sdk.config import *
import json
from dubhe_sdk.service.Logger import Logger
logger = Logger.instance()
from dubhe_sdk.service.service_prepare import *
import traceback

executor = ThreadPoolExecutor(10)
app = Flask(__name__)

model = {}

def getModel(name="entrance_func", defValue=None):
    try:
        return model[name]
    except KeyError:
        return defValue

def setModel(name="entrance_func", value=None):
    model[name] = value

def task_run(dict_params):

    try:
        data = None
        service_type, func, ctxb = getModel()
        assert service_type == dict_params['req_type'], "pod service_type are different from req_type!"

        # send start to kafka
        if dict_params['req_type'] != TASK_INFERENCE_TYPE and PLATFORM_TYPE == AI_PLATFORM:
            start_json = train_start_data()
            send_kafka(MODEL_STATUS, start_json, TOPIC_MODEL_STATUS)

        # set gpu id
        resource_allocation = dict_params['resource_allocation']
        os.environ['CUDA_VISIBLE_DEVICES'] = resource_allocation['CUDA_VISIBLE_DEVICES']

        ctx = ctxb.setInputParam(dict_params).build()
        data = func(ctx)

        # send end to kafka
        if dict_params['req_type'] != TASK_INFERENCE_TYPE and PLATFORM_TYPE == AI_PLATFORM:
            end_json = train_end_data()
            send_kafka(MODEL_STATUS, end_json, TOPIC_MODEL_STATUS)

    except Exception as ex:
        traceback.print_exc()
        exceptionMSG = traceback.format_exc()
        if dict_params['req_type'] != TASK_INFERENCE_TYPE and PLATFORM_TYPE == AI_PLATFORM:
            end_json = train_exception_end_data('%s current model error %s' % (INSTANCE_ID, ex))
            send_kafka(MODEL_STATUS, end_json, TOPIC_MODEL_STATUS)
        logger.error(exceptionMSG)
    finally:
        return data

# region 0. heart detection
@app.route('/heart', methods=['GET', 'POST'])
def heart():
    res = {
            "flag":1,
            "message": "service online"
            }
    return jsonify(res)
# endregion

# region 1. offline train
@app.route('/train/start', methods=['GET', 'POST'])
def train():
    params = flask_request.get_json()
    logger.info("input params:{}".format(params))
    if type(params) == type(dict()):
        dict_params = params
    else:
        dict_params = json.loads(params)
    print('/train/start %s\n' % dict_params)
    # instance_id check
    if dict_params['instance_id'] != INSTANCE_ID:
        res = {
            "flag": 2,
            "message": "param error: instance_id [{}] of request different from ENV INSTANCE_ID [{}]!"
            .format(dict_params['instance_id'], INSTANCE_ID)
        }
        return jsonify(res)

    executor.submit(task_run, dict_params)
    res = {
            "flag":1,
            "message": "model train starting now"
            }
    return jsonify(res)
# endregion

# region 2. offline evaluate
@app.route('/predict/batch', methods=['GET', 'POST'])
def predict_batch():
    params = flask_request.get_json()
    logger.info("input params:{}".format(params))
    if type(params) == type(dict()):
        dict_params = params
    else:
        dict_params = json.loads(params)
    print('/predict/batch %s\n' % dict_params)

    # instance_id check
    if dict_params['instance_id'] != INSTANCE_ID:
        res = {
            "flag": 2,
            "message": "param error: instance_id [{}] of request different from ENV INSTANCE_ID [{}]!"
                .format(dict_params['instance_id'], INSTANCE_ID)
        }
        return jsonify(res)

    executor.submit(task_run, dict_params)
    res = {
            "flag":1,
            "message": "model predict starting now"
            }
    return jsonify(res)
# endregion

# region 3. online inference
@app.route('/predict/multiple', methods=['GET', 'POST'])
def predict_multiple():
    params = flask_request.get_json()
    logger.info("input params:{}".format(params))
    if type(params) == type(dict()):
        dict_params = params
    else:
        dict_params = json.loads(params)
    print('/predict/multiple %s\n' % dict_params)

    # instance_id check
    if dict_params['instance_id'] != INSTANCE_ID:
        res = {
            "flag": 2,
            "message": "param error: instance_id [{}] of request different from ENV INSTANCE_ID [{}]!"
                .format(dict_params['instance_id'], INSTANCE_ID)
        }
        return jsonify(res)

    data = task_run(dict_params)
    res = {
        "flag": 1,
        "message": "model inference success",
        "data": data
    }

    logger.info("response:{}".format(json.dumps(res, indent=4)))
    return jsonify(res)
# endregion


if __name__ == '__main__':
    app.run(host=IP, port=PORT, threaded=THREADED)
    # send ready to kafka



from dubhe_sdk.param.base_param import BaseParam
from dubhe_sdk.config import *


class TrainParam(BaseParam):
    # 请求类型：1.离线预测 2.离线评估 3.在线推理
    req_type = 1
    # 实例ID
    instance_id = INSTANCE_ID
    # 多个数据集路径
    data_paths = []
    # 训练类别
    code_list = []
    # 1：占比   2：指定验证集
    verify_way = 1
    # 验证集占比
    split_rate = 0.1
    # 依赖其他模型路径
    rely_model_data = RELY_MODEL_DATA
    # 资源类型：0.CPU模式  1.GPU模式
    resource_type = 1
    # 资源配置信息
    resource_allocation = {
        "CUDA_VISIBLE_DEVICES": "0"  # None, 0, 1, 2, 3
    }
    # 业务参数
    special_params = special_params
    # 预训练模型路径
    pretrained_model_path = ""
    # 标识断点训练的epoch值，默认0
    pretrained_epoch = 0


from dubhe_sdk.service import (
    train_decorator as train,
    predict_decorator as predict,
    inference_decorator as inference,
    pipeline_decorator as pipeline
)

from dubhe_sdk.service.context import (
    Context,
    ContextKey
)



import dubhe_sdk.service.run as service

service.image_run()



import json
import socketserver
from concurrent.futures import ThreadPoolExecutor
# from dubhe_sdk.config import IP, PORT
from dubhe_sdk.service.service_prepare import serverIsReady, serviceCheck, connect_tcp
from dubhe_sdk.service.Logger import Logger
from dubhe_sdk.service.service_prepare import *

import struct
import traceback

logger = Logger.instance()

model = {}

def getModel(name="entrance_func", defValue=None):
    try:
        return model[name]
    except KeyError:
        return defValue

def setModel(name="entrance_func", value=None):
    model[name] = value

def task_run(dict_params):

    try:
        data = None
        service_type, func, ctxb = getModel()
        assert service_type == dict_params['req_type'], "pod service_type are different from req_type!"

        # send start to kafka
        if dict_params['req_type'] != TASK_INFERENCE_TYPE and PLATFORM_TYPE == AI_PLATFORM:
            start_json = train_start_data()
            send_kafka(MODEL_STATUS, start_json, TOPIC_MODEL_STATUS)

        # set gpu id
        resource_allocation = dict_params['resource_allocation']
        os.environ['CUDA_VISIBLE_DEVICES'] = resource_allocation['CUDA_VISIBLE_DEVICES']

        ctx = ctxb.setInputParam(dict_params).build()
        data = func(ctx)

        # send end to kafka
        if dict_params['req_type'] != TASK_INFERENCE_TYPE and PLATFORM_TYPE == AI_PLATFORM:
            end_json = train_end_data()
            send_kafka(MODEL_STATUS, end_json, TOPIC_MODEL_STATUS)

    except Exception as ex:
        traceback.print_exc()
        exceptionMSG = traceback.format_exc()
        if dict_params['req_type'] != TASK_INFERENCE_TYPE and PLATFORM_TYPE == AI_PLATFORM:
            end_json = train_exception_end_data('%s current model error %s' % (INSTANCE_ID, ex))
            send_kafka(MODEL_STATUS, end_json, TOPIC_MODEL_STATUS)
        logger.error(exceptionMSG)
    finally:
        return data

class TcpHandler(socketserver.BaseRequestHandler):
    def __init__(self, request, client_address, server):
        self.executor = ThreadPoolExecutor(10)
        super(TcpHandler, self).__init__(request, client_address, server)

    def recvMsg(self):
        head = bytes()
        msgBody = bytes()
        while len(head) < 4:
            head += self.request.recv(4-len(head))
            print(head)
            if (len(head) == 0):
                return

        msgLen = struct.unpack('!i', head)[0]
        while msgLen > 0:
            data = self.request.recv(1024)
            msgBody += data
            msgLen -= len(data)
        return msgBody.decode("utf-8")

    def sendMsg(self, msg):
        bMsg = msg.encode('utf-8')
        length = len(bMsg)
        lenData = struct.pack('!i', length)
        self.request.send(lenData)
        self.request.sendall(bMsg)

    def handle(self):
        logger.info('enter---------------handler')
        while True:
            try:
                data = self.recvMsg()
                if not data:
                    break
                logger.info("recv: %s"%data)
                res = self.dispatcher(data)
                self.sendMsg(res)
                logger.info("send: %s"%res)

            except ConnectionResetError as e:
                logger.error("handle msg error", e)
                break

    def dispatcher(self, msg):
        response = {"flag": 1}
        dict_params = None
        try:
            dict_params = json.loads(msg)
            if(not isinstance(dict_params, dict)):
                response.update({"flag": 2, "message": "param error: request is not a dict"})
                return json.dumps(self.response)

            # instance_id check
            if dict_params['instance_id'] != INSTANCE_ID:
                response.update({"flag": 2, "message": "param error: instance_id [{}] of request different from ENV INSTANCE_ID [{}]!"
                                .format(dict_params['instance_id'],INSTANCE_ID)})
                return json.dumps(self.response)

            reqType = dict_params.get('req_type', -1)
            canOffer, content = serviceCheck(dict_params)
            if canOffer:
                if 0 == reqType:
                    # 心跳检测
                    response["message"] = "service online"
                elif 1 == reqType:
                    # 离线训练
                    self.executor.submit(task_run, dict_params)
                    response["message"] = "model train starting now"
                elif 2 == reqType:
                    # 离线评估
                    self.executor.submit(task_run, dict_params)
                    response["message"] = "model predict starting now"
                elif 3 == reqType:
                    # 在线推理
                    response["message"] = "model inference success"
                    response['request_number'] = dict_params['request_number']
                    response['data'] = task_run(dict_params)
                    if not response['data']:
                        response.update({"flag": 2, "message": "exception occurs"})
                else:
                    response.update({"flag": 2, "message": "req_type error"})
            else:
                response.update({"flag": 2, "message": content})
        except ValueError:
            response.update({"flag": 2, "message": "param error"})

        finally:
            return json.dumps(response, indent=4)



class BaseParam(object):

    @classmethod
    def _default_values(cls):
        data = {key: cls.__dict__[key] for key in cls.__dict__.keys() if not key.startswith('_')}
        return data

    def __str__(self):
        inKeys = set([key for key in self.__dict__.keys() if not key.startswith('__')])
        clzKeys = set([key for key in self.__class__.__dict__.keys() if not key.startswith('__')])
        keys = inKeys.union(clzKeys)
        out = ''
        for key in keys:
            if key in self.__dict__:
                out += '%s:%s\n' % (key, self.__dict__[key])
            else:
                out += '%s:%s\n' % (key, self.__class__.__dict__[key])
        return out


import os
import logging
from logging.handlers import RotatingFileHandler

class Logger(object):
    logger = None

    def __init__(self, loglevel = logging.INFO, maxMB = 10, backupCount = 14):
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls):
        if cls.logger is None:
            #TODO: config to limit logs by days
            path = "/temporary/log"

            if not os.path.exists(path):
                os.makedirs(path)
            filename = '%s/dubhe.log'%path
            formatter = logging.Formatter('%(asctime)s | %(levelname)-8s | %(filename)s:%(lineno)s: %(message)s')
            # 控制台输出
            streamHandler = logging.StreamHandler()
            streamHandler.setFormatter(formatter)
            # 滚动文件输出
            log_file_handler = RotatingFileHandler(filename=filename, mode='a', maxBytes=10*1024*1024, backupCount=14, encoding=None, delay=0)
            log_file_handler.setFormatter(formatter)
            cls.logger = logging.getLogger()
            cls.logger.setLevel(logging.INFO)
            cls.logger.addHandler(streamHandler)
            cls.logger.addHandler(log_file_handler)

            streamHandler.close()
            log_file_handler.close()

            # 过滤kafka INFO级别日志
            kafka_logger = logging.getLogger('kafka')
            kafka_logger.setLevel(logging.WARNING)

        return cls.logger


from dubhe_sdk.config import *
from dubhe_sdk.service.ADCkafka import *
from dubhe_sdk.service import context
from dubhe_sdk.service.service_prepare import *
from concurrent.futures import ThreadPoolExecutor
import socketserver

import sys
import importlib
import random
import typing
import inspect
import traceback


def image_run():
    try:
        serve_flag = False
        main_file_path = sys.argv[0]
        pipelines = _find_pipelines(main_file_path)
        entrances = _find_entrances(pipelines)

        # 中间结果目录
        TMP_DIR = '/temporary/debug'

        ctxb = context.ContextBuilder(None)\
            .add(context.ContextKey.ENV_INSTANCE_ID, INSTANCE_ID)\
            .add(context.ContextKey.ENV_PLATFORM_TYPE, PLATFORM_TYPE)\
            .add(context.ContextKey.ENV_RELY_MODEL_DATA, RELY_MODEL_DATA)\
            .add(context.ContextKey.ENV_MODEL_PATH, model_path)\
            .add(context.ContextKey.ENV_SPECIAL_PARAMS, special_params)\
            .add(context.ContextKey.KEY_TMP_DIR, TMP_DIR)
        ctx = ctxb.build()

        for entrance in entrances:
            func_type = entrance[2]
            # 训练：modelid1412-algorithm112345540-train-1412-4
            # 推理：taskid84-application170644901-inference-84
            if func_type == TASK_TYPE:
                module = entrance[0]
                name = entrance[1]
                user_Model = module.__getattribute__(name)
                Model = user_Model(ctx)
                function = entrance[3]
                user_func = function.__get__(Model)
                model_infos = (func_type, user_func, ctxb)

                if 'protocol' in os.environ and str(os.environ['protocol']) == "http":
                    # HTTP通信方式
                    from dubhe_sdk.http_server import app, setModel
                    setModel(value=model_infos)
                    ThreadPoolExecutor(1).submit(open_browser)
                    app.run(host=IP, port=PORT, threaded=THREADED)
                else:
                    # TCP通信方式
                    from dubhe_sdk.tcp_server import TcpHandler, setModel
                    setModel(value=model_infos)
                    ThreadPoolExecutor(1).submit(connect_tcp, IP, PORT)
                    server = socketserver.ThreadingTCPServer((IP, PORT), TcpHandler)
                    server.serve_forever()
                serve_flag = True
        if serve_flag == False:
            raise Exception('Serve not found!')

    except Exception as ex:
        traceback.print_exc()
        exceptionMSG = traceback.format_exc()
        if TASK_TYPE != TASK_INFERENCE_TYPE and PLATFORM_TYPE == AI_PLATFORM:
            end_json = train_exception_end_data('%s current model error %s' % (INSTANCE_ID, ex))
            send_kafka(MODEL_STATUS, end_json, TOPIC_MODEL_STATUS)
        logger.error(exceptionMSG)


def _find_pipelines(filename: str) -> typing.List[typing.Tuple]:
    pipelines = []

    assert filename[-3:] == '.py' and os.path.isfile(filename), 'Python entry point is not correct'

    _, module_name_with_ext = os.path.split(filename)
    module_name, _ = os.path.splitext(module_name_with_ext)
    module_spec = importlib.util.spec_from_file_location(module_name, filename)
    module = importlib.util.module_from_spec(module_spec)
    module_spec.loader.exec_module(module)
    for name in dir(module):
        if name.startswith('__'):
            continue

        obj = module.__getattribute__(name)
        if hasattr(obj, '_is_pipeline_func'):
            pipelines.append((module, name))
    return pipelines

def _find_entrances(pipelines: typing.List[typing.Tuple]) -> typing.List[typing.Tuple]:
    entrances = []

    for pipe in pipelines:
        module = pipe[0]
        name = pipe[1]
        obj = module.__getattribute__(name)
        entrance = None
        for _, function in inspect.getmembers(
                obj,
                predicate=lambda x: inspect.isfunction(x) or inspect.ismethod(x),
        ):
            func_type = None
            if hasattr(function, "_is_train_func"):
                func_type = TASK_TRAIN_TYPE
            elif hasattr(function, "_is_predict_func"):
                func_type = TASK_PREDICT_TYPE
            elif hasattr(function, '_is_inference_func'):
                func_type = TASK_INFERENCE_TYPE

            if func_type is not None:
                entrance = (module, name, func_type, function)
                entrances.append(entrance)

    return entrances

# def taskType():
#     curType = INSTANCE_ID.split('-')[2]
#     if curType == PREFIX_TRAIN:
#         return TASK_TRAIN_TYPE
#     elif curType == PREFIX_PREDICT:
#         return TASK_PREDICT_TYPE
#     elif curType == PREFIX_INFERENCE or curType == PREFIX_AUTOMARK:
#         return TASK_INFERENCE_TYPE
#     else:
#         logger.error("TASK %s PREFIX not defined!"%curType)
#         return -1


import typing
from dubhe_sdk.dataset import dataset

class Dataset(object):
    def __init__(self, ds: dataset.Dataset):
        self._org_dataset = ds

    def getCount(self):
        return self._org_dataset.getCount()

    def getId(self, index: int) -> str:
        return self._org_dataset.getId(index)

    def openFile(self, index: int, key: str = "default") -> typing.IO:
        """
        Return file of a dataset if has

        Args:
            index: index of item
            key: Key of the file, if there is only one image in each item, use 'default'
            or you must assign them different key in dubhe cloud

        :return: file: the file object of the item
        """
        return self._org_dataset.openFile(index, key)

    def getClassification(self, index: int, key: str = "default") -> str:
        return self._org_dataset.getClassification(index, key)

    def getDetection(self, index: int, key: str = "default") -> typing.List:
        return self._org_dataset.getDetection(index, key)

    def getMetadata(self, index: int, key: str = "default") -> str:
        return self._org_dataset.getMetadata(index, key)


# import time
from dubhe_sdk.service.ADCkafka import *
import urllib.request as request
# send pretending msg
import socket
from dubhe_sdk.config import *

from dubhe_sdk.service.Logger import Logger
logger = Logger.instance()


def serviceCheck(dict_params):

    # 1:gpu , 0:cpu
    resource_type = dict_params['resource_type']
    # req_type = 0: heart, 1: train,2: offline inference, 3: online inference
    req_type = dict_params['req_type']
    # train cpu is not support now req_type=1
    if req_type == 1 and resource_type == 0:
        return False, 'train cpu is not support now'

    return True, 'sevice online'


def serverIsReady():
    print("enter serverIsReady")
    if KAFKA_ENABLE:
        print("prepara ready flag.")
        ready_json = model_ready_data()
        send_kafka(MODEL_READY, ready_json, TOPIC_MODEL_STATUS)
        print("ready send success!")



def open_browser():
    logger.info('server starting...')
    while True:
        try:
            request.urlopen(url=INDEX_URL)
            break
        except Exception as e:
            print(e)
            time.sleep(0.5)
    if (PLATFORM_TYPE == AI_PLATFORM):
        serverIsReady()
    logger.info('server started !')

def connect_tcp(host, port):
    logger.info('server starting...')
    while True:
        # logger.info("****")
        try:
            client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            code = client.connect_ex((host,port))
            client.close()
            if code == 0:
                break
        except Exception as e:
            print(e)
            time.sleep(1)
    if (PLATFORM_TYPE == AI_PLATFORM):
        serverIsReady()
    logger.info('server started !')





from dubhe_sdk.param.base_param import BaseParam
from dubhe_sdk.config import *


class PredictParam(BaseParam):
    # 请求类型：1.离线预测 2.离线评估 3.在线推理
    req_type = 2
    # 实例ID
    instance_id = INSTANCE_ID
    # 多个数据集路径
    data_paths = []
    # 模型文件路径
    model_path = model_path
    # 依赖其他模型路径
    rely_model_data = RELY_MODEL_DATA
    # 资源模式：0.CPU模式  1.GPU模式
    resource_type = 1
    # 资源配置信息
    resource_allocation = {
        "CUDA_VISIBLE_DEVICES": "0"  # None, 0, 1, 2, 3
    }
    # 业务参数
    special_params = special_params
    # 中间结果目录
    BASE_DIR = '/temporary/debug'
    # 模型输出目录
    OUTPUT_DIR = 'output'


from setuptools import setup, find_packages

setup(name='dubhe_sdk',
      version="2.13",
      description='TCL AI platform, algorithm software development kit',
      long_description='TCL AI platform, algorithm software development kit',
      classifiers=[],
      keywords='dubhe',
      author='huanghui',
      author_email='',
      url='https://github.com/hathawayhh/dubhe-sdk',
      license='MIT',
      packages=find_packages(),
      data_files=[],
      include_package_data=True,
      zip_safe=True,
      install_requires=[
        "Flask==1.1.1",
        "kafka-python==2.0.1"
      ])




from kafka import KafkaProducer
import json
from dubhe_sdk.config import *
import uuid
import time
import datetime

def msg_id():
    return uuid.uuid1().int>>64

# 模型状态信息
def model_ready_data():
    """
    {
		"action": "READY" # START, END ,
		"status": 200 # 200: 正常 /300: 异常,
		"msg": "content"
	}
    :return:
    """
    data = {}
    data['action'] = 'READY'
    data['status'] = 200
    data['msg'] = 'model is ready'
    return data

def train_start_data():
    """
    {
		"action": "START" # START, END ,
		"status": 200 # 200: 正常 /300: 异常,
		"msg": "content"
	}
    :return:
    """
    data = {}
    data['action'] = 'START'
    data['status'] = 200
    data['msg'] = 'model train starting'
    return data

def train_end_data():
    """
    {
		"action": "START" # START, END ,
		"status": 200 # 200: 正常 /300: 异常,
		"msg": "content"
	}
    :return:
    """
    data = {}
    data['action'] = 'END'
    data['status'] = 200
    data['msg'] = 'model train end'
    return data

def train_exception_end_data(exceptionMSG):
    """
    {
		"action": "START" # START, END ,
		"status": 200 # 200: 正常 /300: 异常,
		"msg": "content"
	}
    :return:
    """
    data = {}
    data['action'] = 'END'
    data['status'] = 300
    data['msg'] = exceptionMSG
    return data

def predict_start_data():
    """
    {
		"action": "START" # START, END ,
		"status": 200 # 200: 正常 /300: 异常,
		"msg": "content"
	}
    :return:
    """
    data = {}
    data['action'] = 'START'
    data['status'] = 200
    data['msg'] = 'model predict starting'
    return data

def predict_end_data():
    """
    {
		"action": "START" # START, END ,
		"status": 200 # 200: 正常 /300: 异常,
		"msg": "content"
	}
    :return:
    """
    data = {}
    data['action'] = 'END'
    data['status'] = 200
    data['msg'] = 'model predict end'
    return data

def predict_exception_end_data(exceptionMSG):
    """
    {
		"action": "START" # START, END ,
		"status": 200 # 200: 正常 /300: 异常,
		"msg": "content"
	}
    :return:
    """
    data = {}
    data['action'] = 'END'
    data['status'] = 300
    data['msg'] = exceptionMSG
    return data

# 模型进度信息
def train_progress_data(progress, epoch, metrics=dict()):
    """
    {
        "progress": 0.2 # 0-1 float ,
        "epoch": 0 # 0 - total epochs,
        "metrics": {
                "loss": 1.1 # float,
                "lr": 0.001 # float ,
                "iou": # float,
                "acc": 0.5 # 0-1 float,
                "recall": 0.5 # 0-1 float
            } # float,
	}
    :param progress:
    :param epoch:
    :param metrics:
    :return:
    """

    data = metrics
    data['progress'] = progress
    data['epoch'] = epoch
    return data

def predict_progress_data(progress):
    """
    {
        "progress": 0.2 # 0-1 float ,
	}
    :param progress:
    :return:
    """
    data = {}
    data['progress'] = progress
    return data

def predict_classify_details(img_path, ground_true, predict_code, conf):
    """
    {
		"img_path": "img_full_path"  ,
		"ground_true": "code1" # ,
		"predict_code": "code2",
		"conf": 0.8 # 0-1 float ,
	}
    :return:
    """
    data_list = []
    data = {}
    data['ground_true'] = ground_true
    data['predict_code'] = predict_code
    data['conf'] = conf
    data_list.append(data)

    result = {}
    result['img_path'] = img_path
    result['info'] = data_list
    return [result]

def predict_details(img_path, ground_true, predict_code, conf, shape_type=None,position=None,special=None):
    """
	{
		"img_path": "img_full_path"  ,
		"pre_img": "preimg_full_path"  ,
		"ground_true": "code1" # ,
		"predict_code": "code2",
		"conf": 0.8 # 0-1 float ,
	}
    :return:
    """
    data_list = []
    if not position:
        data = {}
        data['ground_true'] = ground_true
        data['predict_code'] = predict_code
        data['conf'] = conf
    elif len(position) > 0:
        for pos in position:
            (x1, y1, x2, y2, boxconf) = pos
            data = {}
            data['ground_true'] = ground_true
            data['predict_code'] = predict_code
            data['conf'] = conf
            if conf<0:
                data['conf'] = boxconf
            data['shape_type'] = shape_type
            data['position'] = [[x1,y1], [x2,y2]]
            data['special'] = special
            data_list.append(data)
    else:
        data = {}
        data['ground_true'] = ground_true
        data['predict_code'] = predict_code
        data['conf'] = conf
        if conf < 0:
            data['conf'] = 1
        data['shape_type'] = shape_type
        data['position'] = []
        data['special'] = special

    data_list.append(data)
    result = {}
    result['img_path'] = img_path
    if position:
        result['bbox_cnt'] = len(position)
    result['info'] = data_list
    return [result]


def msg_context(mes_type, data):
    kafka_json = {}
    kafka_json['instance_id'] = KafkaParam.instance_id
    kafka_json['request_number'] = msg_id()
    kafka_json['date'] = int(time.time()*1000)
    kafka_json['mes_type'] = mes_type
    kafka_json['data'] = data
    return kafka_json

def send_kafka(mes_type, data, topic=topic_name1, local_backup='json.dat'):
    """
    :param topic:
    :param message: json format
    :return:
    """
    msg = msg_context(mes_type, data)
    time_str = datetime.datetime.strftime(datetime.datetime.now(), '%Y-%m-%d %H:%M:%S')
    with open(local_backup, encoding='utf-8', mode='a+') as writer:
        writer.write('%s, %s, %s\n' % (time_str, topic, msg))
    if KAFKA_ENABLE:
        try:
            producer = KafkaProducer(bootstrap_servers=BOOTSTRAP_SERVERS,
                                     value_serializer=lambda m: json.dumps(m).encode('utf-8'))
            producer.send(topic, value=msg)
            producer.close()
        except Exception as e:
            print(msg)
            print(e)
            raise e

class KafkaParam(object):
    instance_id = INSTANCE_ID




# -*- coding=utf-8 -*-
"""
    Desc:
    Auth: LiZhifeng
    Date: 2020/6/9
"""
import os
import uuid
import json
from dubhe_sdk.service.Logger import Logger

logger = Logger.instance()

isDEBUG = False

logger.info('**************config.py*****************')

DEFAULT_TOPIC_MODEL_STATUS = 'alg_model_progress'
DEFAULT_TOPIC_MODEL_RESULT = 'als_model_result'
IP = "0.0.0.0"
# kafka配置
KAFKA_ENABLE = True
THREADED = True
# TOPIC配置
topic_name1 = "alg_train_progress"

# 启动配置
DEBUG_PLATFORM = 0
AI_PLATFORM = 1

if 'PLATFORM_TYPE' in os.environ and int(os.environ['PLATFORM_TYPE']) == AI_PLATFORM:
    logger.info('AI Platfrom Started Algorithm!')
    logger.info('---------------ENV------------------')
    PLATFORM_TYPE = AI_PLATFORM

    PORT = int(os.environ['PORT'])
    logger.info('PORT %s ' % os.environ['PORT'])

    BOOTSTRAP_SERVERS = os.environ['BOOTSTRAP_SERVERS']
    logger.info('BOOTSTRAP_SERVERS %s ' % BOOTSTRAP_SERVERS)

    TOPIC_MODEL_STATUS = os.environ['TOPIC_MODEL_STATUS']
    logger.info('TOPIC_MODEL_STATUS %s' % TOPIC_MODEL_STATUS)

    TOPIC_MODEL_RESULT = os.environ['TOPIC_MODEL_RESULT']
    logger.info('TOPIC_MODEL_RESULT  %s ' % TOPIC_MODEL_RESULT)

    INSTANCE_ID = os.environ['INSTANCE_ID']
    logger.info('INSTANCE_ID %s' % INSTANCE_ID)

    INDEX_URL = "http://%s:%d/heart" % (os.environ['INSTANCE_ID'], PORT)
    logger.info('INDEX_URL %s ' % INDEX_URL)

    RELY_MODEL_DATA = None
    if 'RELY_MODEL_DATA' in os.environ and os.environ['RELY_MODEL_DATA'] != None and len(os.environ['RELY_MODEL_DATA']) != 0:
        RELY_MODEL_DATA = json.loads(os.environ['RELY_MODEL_DATA'])
        logger.info('RELY_MODEL_DATA %s ' % RELY_MODEL_DATA)

    special_params = None
    if 'special_params' in os.environ and os.environ['special_params'] != None and len(os.environ['special_params']) != 0:
        special_params = json.loads(os.environ['special_params'])
        logger.info('special_params %s ' % special_params)

    model_path = None
    if 'model_path' in os.environ and os.environ['model_path'] != None:
        model_path = os.environ['model_path']
        logger.info('model_path %s ' % model_path)


    logger.info('---------------ENV------------------')
else:
    logger.info('Manual Started Algoritm!')
    logger.info('---------------DEFAULT------------------')
    PLATFORM_TYPE = DEBUG_PLATFORM
    IP = "127.0.0.1"
    PORT = 8030
    logger.info('Default IP: %s, PORT: %d' %(IP,PORT))

    # BOOTSTRAP_SERVERS = ['192.168.7.114:9092', '192.168.7.113:9092']
    BOOTSTRAP_SERVERS = '127.0.0.1:9092'
    logger.info('Default BOOTSTRAP_SERVERS, use %s' % BOOTSTRAP_SERVERS)

    TOPIC_MODEL_STATUS = DEFAULT_TOPIC_MODEL_STATUS
    logger.info('Default TOPIC_MODEL_STATUS, use default %s ' % TOPIC_MODEL_STATUS)

    TOPIC_MODEL_RESULT = DEFAULT_TOPIC_MODEL_RESULT
    logger.info('Default TOPIC_MODEL_RESULT, use default %s ' % TOPIC_MODEL_RESULT)

    INSTANCE_ID = "modelid1412-algorithm112345540-train-1412-4"
    # INSTANCE_ID = "forecastid323-algorithm154146065-forecast-323-4"
    # INSTANCE_ID = "taskid14-application203913747-inference-14"
    logger.info('Default INSTANCE_ID, use %s ' % INSTANCE_ID)

    INDEX_URL = "http://127.0.0.1:%d/heart" % PORT
    logger.info('Default INDEX_URL, use {} '.format(INDEX_URL))

    # RELY_MODEL_DATA = {
    #     "detector": {
    #         "alg_type": "目标检测",
    #         "alg_version": "v1.0",
    #         "model_path": "/temporary/debug/model_20201029082410454_1_0.51.pth",
    #         "model_version": "v7.0"
    #     },
    #     "classify": {
    #         "alg_type": "分类",
    #         "alg_version": "v1.0",
    #         "model_path": "/temporary/debug/model_20201029084953456_4_2.29.pth",
    #         "model_version": "v7.0"
    #     }
    # }

    # RELY_MODEL_DATA = {
    #     "unet": {
    #         "alg_type": "目标检测",
    #         "alg_version": "v1.0",
    #         "model_path": "./ckpts/Unet_se_resnext50_32x4d_lowest_loss.pth",
    #         "model_version": "v7.0"
    #     }
    # }

    RELY_MODEL_DATA = {
        "classification_model": {
            "alg_type": "目标检测",
            "alg_version": "v1.0",
            "model_path": "./output/models/model_20210830140750571_1_0.00.pth",
            "model_version": "v7.0"
        }
    }

    logger.info('Default RELY_MODEL_DATA, use %s ' % RELY_MODEL_DATA)

    special_params = {
        "section_id": "1E953",  # 站点ID,
        "debug_model": 0,  # 0:关闭, 1:开启
    }
    # model_path = None
    model_path = "./output/models/model_20210830140750571_1_0.00.pth"


    logger.info('---------------DEFAULT------------------')


# message type
# status
MODEL_READY = 0
MODEL_STATUS = 1
MODEL_PROGRESS = 2

#result type
MODEL_RESULT = 3

# instance id prefix
PREFIX_TRAIN = 'train'
PREFIX_PREDICT = 'forecast'
PREFIX_INFERENCE = 'inference'
PREFIX_AUTOMARK = 'automark'

# task type
TASK_TRAIN_TYPE = 1
TASK_PREDICT_TYPE = 2
TASK_INFERENCE_TYPE = 3

if INSTANCE_ID.split('-')[2] == PREFIX_TRAIN:
    TASK_TYPE = TASK_TRAIN_TYPE
elif INSTANCE_ID.split('-')[2] == PREFIX_PREDICT:
    TASK_TYPE =  TASK_PREDICT_TYPE
elif INSTANCE_ID.split('-')[2] == PREFIX_INFERENCE or INSTANCE_ID.split('-')[2] == PREFIX_AUTOMARK:
    TASK_TYPE =  TASK_INFERENCE_TYPE
else:
    logger.error("TASK %s PREFIX not defined!"%INSTANCE_ID.split('-')[2])
    TASK_TYPE =  -1
